@@ -1,7 +1,7 @@
 /* pptp_ctrl.c ... handle PPTP control connection.
  *                 C. Scott Ananian <cananian@alumni.princeton.edu>
  *
- * $Id: pptp_ctrl.c,v 1.25 2004/06/10 07:11:57 quozl Exp $
+ * $Id: pptp_ctrl.c,v 1.31 2005/03/31 07:42:39 quozl Exp $
  */
 
 #include <errno.h>
@@ -165,20 +165,24 @@ static const char *pptp_start_ctrl_conn_rply[] = {
 
 #define MAX_START_CTRL_CONN_REPLY 4
 
+/* timing options */
+int idle_wait = PPTP_TIMEOUT;
+int max_echo_wait = PPTP_TIMEOUT;
+
 /* Local prototypes */
 static void pptp_reset_timer(void);
-static void pptp_handle_timer(int sig);
+static void pptp_handle_timer();
 /* Write/read as much as we can without blocking. */
-void pptp_write_some(PPTP_CONN * conn);
-void pptp_read_some(PPTP_CONN * conn);
+int pptp_write_some(PPTP_CONN * conn);
+int pptp_read_some(PPTP_CONN * conn);
 /* Make valid packets from read_buffer */
 int pptp_make_packet(PPTP_CONN * conn, void **buf, size_t *size);
 /* Add packet to write_buffer */
 int pptp_send_ctrl_packet(PPTP_CONN * conn, void * buffer, size_t size);
 /* Dispatch packets (general) */
-void pptp_dispatch_packet(PPTP_CONN * conn, void * buffer, size_t size);
+int pptp_dispatch_packet(PPTP_CONN * conn, void * buffer, size_t size);
 /* Dispatch packets (control messages) */
-void ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size);
+int ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size);
 /* Set link info, for pptp servers that need it.
    this is a noop, unless the user specified a quirk and
    there's a set_link hook defined in the quirks table
@@ -237,6 +241,10 @@ static void ctrlp_rep( void * buffer, int size, int isbuff)
     unsigned int type;
     if(size < sizeof(struct pptp_header)) return;
     type = ntoh16(packet->ctrl_type);
+    /* FIXME: do not report sending echo requests as long as they are
+     * sent in a signal handler. This may dead lock as the syslog call
+     * is not reentrant */
+    if( type ==  PPTP_ECHO_RQST ) return;
     /* don't keep reporting sending of echo's */
     if( (type == PPTP_ECHO_RQST || type == PPTP_ECHO_RPLY) && nlogecho <= 0 ) return;
     log("%s control packet type is %d '%s'\n",isbuff ? "Buffered" : "Sent", 
@@ -249,7 +257,6 @@ static void ctrlp_rep( void * buffer, int size, int isbuff)
 /* Open new pptp_connection.  Returns NULL on failure. */
 PPTP_CONN * pptp_conn_open(int inet_sock, int isclient, pptp_conn_cb callback)
 {
-    struct sigaction sigact;
     PPTP_CONN *conn;
     /* Allocate structure */
     if ((conn = malloc(sizeof(*conn))) == NULL) return NULL;
@@ -298,11 +305,9 @@ PPTP_CONN * pptp_conn_open(int inet_sock, int isclient, pptp_conn_cb callback)
     }
     /* Set up interval/keep-alive timer */
     /*   First, register handler for SIGALRM */
-    sigact.sa_handler = pptp_handle_timer;
-    sigemptyset(&sigact.sa_mask);
-    sigact.sa_flags = SA_RESTART;
+    sigpipe_create();
+    sigpipe_assign(SIGALRM);
     global.conn = conn;
-    sigaction(SIGALRM, &sigact, &global.old_sigaction);
     /* Reset event timer */
     pptp_reset_timer();
     /* all done. */
@@ -439,7 +444,7 @@ void pptp_conn_destroy(PPTP_CONN * conn)
         pptp_call_destroy(conn, vector_get_Nth(conn->call, i));
     /* notify */
     if (conn->callback != NULL) conn->callback(conn, CONN_CLOSE_DONE);
-    sigaction(SIGALRM, &global.old_sigaction, NULL);
+    sigpipe_close();
     close(conn->inet_sock);
     /* deallocate */
     vector_destroy(conn->call);
@@ -459,54 +464,67 @@ void pptp_fd_set(PPTP_CONN * conn, fd_set * read_set, fd_set * write_set,
     /* Always add fd to read_set. (always want something to read) */
     FD_SET(conn->inet_sock, read_set);
     if (*max_fd < conn->inet_sock) *max_fd = conn->inet_sock;
+    /* Add signal pipe file descriptor to set */
+    int sig_fd = sigpipe_fd();
+    FD_SET(sig_fd, read_set);
+    if (*max_fd < sig_fd) *max_fd = sig_fd;
 }
 
 /*** handle any pptp file descriptors set in fd_set, and clear them ***********/
-void pptp_dispatch(PPTP_CONN * conn, fd_set * read_set, fd_set * write_set)
+int pptp_dispatch(PPTP_CONN * conn, fd_set * read_set, fd_set * write_set)
 {
+    int r = 0;
     assert(conn && conn->call);
+    /* Check for signals */
+    if (FD_ISSET(sigpipe_fd(), read_set)) {
+        if (sigpipe_read() == SIGALRM) pptp_handle_timer();
+	FD_CLR(sigpipe_fd(), read_set);
+    }
     /* Check write_set could be set. */
     if (FD_ISSET(conn->inet_sock, write_set)) {
         FD_CLR(conn->inet_sock, write_set);
         if (conn->write_size > 0)
-            pptp_write_some(conn);/* write as much as we can without blocking */
+            r = pptp_write_some(conn);/* write as much as we can without blocking */
     }
     /* Check read_set */
-    if (FD_ISSET(conn->inet_sock, read_set)) {
+    if (r >= 0 && FD_ISSET(conn->inet_sock, read_set)) {
         void *buffer; size_t size;
         FD_CLR(conn->inet_sock, read_set);
-        pptp_read_some(conn); /* read as much as we can without blocking */
+        r = pptp_read_some(conn); /* read as much as we can without blocking */
+	if (r < 0)
+	    return r;
         /* make packets of the buffer, while we can. */
-        while (pptp_make_packet(conn, &buffer, &size)) {
-            pptp_dispatch_packet(conn, buffer, size);
+        while (r >= 0 && pptp_make_packet(conn, &buffer, &size)) {
+            r = pptp_dispatch_packet(conn, buffer, size);
             free(buffer);
         }
     }
     /* That's all, folks.  Simple, eh? */
+    return r;
 }
 
 /*** Non-blocking write *******************************************************/
-void pptp_write_some(PPTP_CONN * conn) {
+int pptp_write_some(PPTP_CONN * conn) {
     ssize_t retval;
     assert(conn && conn->call);
     retval = write(conn->inet_sock, conn->write_buffer, conn->write_size);
     if (retval < 0) { /* error. */
         if (errno == EAGAIN || errno == EINTR) { 
-            /* ignore */;
+            return 0;
         } else { /* a real error */
             log("write error: %s", strerror(errno));
-            pptp_conn_destroy(conn); /* shut down fast. */
+	    return -1;
         }
-        return;
     }
     assert(retval <= conn->write_size);
     conn->write_size -= retval;
     memmove(conn->write_buffer, conn->write_buffer + retval, conn->write_size);
     ctrlp_rep(conn->write_buffer, retval, 0);
+    return 0;
 }
 
 /*** Non-blocking read ********************************************************/
-void pptp_read_some(PPTP_CONN * conn)
+int pptp_read_some(PPTP_CONN * conn)
 {
     ssize_t retval;
     assert(conn && conn->call);
@@ -514,7 +532,7 @@ void pptp_read_some(PPTP_CONN * conn)
         char *new_buffer = realloc(conn->read_buffer, 
                 sizeof(*(conn->read_buffer)) * conn->read_alloc * 2);
         if (new_buffer == NULL) {
-            log("Out of memory"); return;
+            log("Out of memory"); return -1;
         }
         conn->read_alloc *= 2;
         conn->read_buffer = new_buffer;
@@ -523,22 +541,19 @@ void pptp_read_some(PPTP_CONN * conn)
             conn->read_alloc  - conn->read_size);
     if (retval == 0) {
         log("read returned zero, peer has closed");
-        pptp_conn_destroy(conn); /* shut down fast. */
-        /* BUG: conn is used by callmgr_main() after being freed here */
-        return;
+        return -1;
     }
     if (retval < 0) {
         if (errno == EINTR || errno == EAGAIN)
-            /* ignore */ ;
+	    return 0;
         else { /* a real error */
             log("read error: %s", strerror(errno));
-            pptp_conn_destroy(conn); /* shut down fast. */
-            /* BUG: conn is used by callmgr_main() after being freed here */
+            return -1;
         }
-        return;
     }
     conn->read_size += retval;
     assert(conn->read_size <= conn->read_alloc);
+    return 0;
 }
 
 /*** Packet formation *********************************************************/
@@ -629,15 +644,16 @@ int pptp_send_ctrl_packet(PPTP_CONN * conn, void * buffer, size_t size)
 }
 
 /*** Packet Dispatch **********************************************************/
-void pptp_dispatch_packet(PPTP_CONN * conn, void * buffer, size_t size)
+int pptp_dispatch_packet(PPTP_CONN * conn, void * buffer, size_t size)
 {
+    int r = 0;
     struct pptp_header *header = (struct pptp_header *)buffer;
     assert(conn && conn->call); assert(buffer);
     assert(ntoh32(header->magic) == PPTP_MAGIC);
     assert(ntoh16(header->length) == size);
     switch (ntoh16(header->pptp_type)) {
         case PPTP_MESSAGE_CONTROL:
-            ctrlp_disp(conn, buffer, size);
+            r = ctrlp_disp(conn, buffer, size);
             break;
         case PPTP_MESSAGE_MANAGE:
             /* MANAGEMENT messages aren't even part of the spec right now. */
@@ -648,6 +664,7 @@ void pptp_dispatch_packet(PPTP_CONN * conn, void * buffer, size_t size)
                     (unsigned int) ntoh16(header->pptp_type));
             break;
     }
+    return r;
 }
 
 /*** log echo request/replies *************************************************/
@@ -663,7 +680,7 @@ static void logecho( int type)
 }
 
 /*** pptp_dispatch_ctrl_packet ************************************************/
-void ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size)
+int ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size)
 {
     struct pptp_header *header = (struct pptp_header *)buffer;
     u_int8_t close_reason = PPTP_STOP_NONE;
@@ -674,7 +691,7 @@ void ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size)
     if (size < PPTP_CTRL_SIZE(ntoh16(header->ctrl_type))) {
         log("Invalid packet received [type: %d; length: %d].",
                 (int) ntoh16(header->ctrl_type), (int) size);
-        return;
+        return 0;
     }
     switch (ntoh16(header->ctrl_type)) {
         /* ----------- STANDARD Start-Session MESSAGES ------------ */
@@ -773,19 +790,18 @@ void ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size)
                 if (conn->callback != NULL)
                     conn->callback(conn, CONN_CLOSE_RQST);
                 conn->conn_state = CONN_IDLE;
-                pptp_conn_destroy(conn);
+		return -1;
             }
             break;
         }
         case PPTP_STOP_CTRL_CONN_RPLY:
         {
-            log("Received Stop Control Connection Reqply.");
+            log("Received Stop Control Connection Reply.");
             /* conn_state should be CONN_WAIT_STOP_REPLY, but it 
              * could be something else */
             if (conn->conn_state == CONN_IDLE) break;
             conn->conn_state = CONN_IDLE;
-            pptp_conn_destroy(conn);
-            break;
+	    return -1;
         }
             /* ----------- STANDARD Echo/Keepalive MESSAGES ------------ */
         case PPTP_ECHO_RPLY:
@@ -945,10 +961,11 @@ void ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size)
             /* goto pptp_conn_close; */
             break;
     }
-    return;
+    return 0;
 pptp_conn_close:
     warn("pptp_conn_close(%d)", (int) close_reason);
     pptp_conn_close(conn, close_reason);
+    return 0;
 }
 
 /*** pptp_set_link **************************************************************/
@@ -1009,13 +1026,13 @@ void * pptp_conn_closure_get(PPTP_CONN * conn)
 static void pptp_reset_timer(void)
 {
     const struct itimerval tv = { {  0, 0 },   /* stop on time-out */
-        { PPTP_TIMEOUT, 0 } }; /* 60-second timer */
-        setitimer(ITIMER_REAL, &tv, NULL);
+        { idle_wait, 0 } };
+    if (idle_wait) setitimer(ITIMER_REAL, &tv, NULL);
 }
 
 
 /*** Handle keep-alive timer **************************************************/
-static void pptp_handle_timer(int sig)
+static void pptp_handle_timer()
 {
     int i;
     /* "Keep Alives and Timers, 1": check connection state */
@@ -1027,16 +1044,15 @@ static void pptp_handle_timer(int sig)
             pptp_conn_close(global.conn, PPTP_STOP_NONE);
     }
     /* "Keep Alives and Timers, 2": check echo status */
-    if (global.conn->ka_state == KA_OUTSTANDING) /*no response to keep-alive*/
-        pptp_conn_close(global.conn, PPTP_STOP_NONE);
-    else { /* ka_state == NONE */ /* send keep-alive */
+    if (global.conn->ka_state == KA_OUTSTANDING) {
+        /* no response to keep-alive */
+        log ("closing control connection due to missing echo reply");
+	pptp_conn_close(global.conn, PPTP_STOP_NONE);
+    } else { /* ka_state == NONE */ /* send keep-alive */
         struct pptp_echo_rqst rqst = {
             PPTP_HEADER_CTRL(PPTP_ECHO_RQST), hton32(global.conn->ka_id) };
         pptp_send_ctrl_packet(global.conn, &rqst, sizeof(rqst));
         global.conn->ka_state = KA_OUTSTANDING;
-        /* XXX FIXME: wake up ctrl thread -- or will the SIGALRM do that
-         * automagically? XXX
-         */
     }
     /* check incoming/outgoing call states for !IDLE && !ESTABLISHED */
     for (i = 0; i < vector_size(global.conn->call); i++) {
